;undump -- demo of the ReadD1 subroutine.
;To see what it's about, try these command lines:
;undump "Hello" d a
;undump 'Hello' d a
;undump " o'clock "
;undump 7
;undump 48656C6C6F203030370D0A
;undump 0D0A"Hello"0D0A
;The output from undump.exe can be redirected to a file, but then you
;should avoid the special symbols like |, >, &, and any embedded
;quotation marks, using the hex equivalents instead.
;Sample build file:
; \masm32\bin\ml /c /coff undump.asm
; \masm32\bin\polink /SUBSYSTEM:CONSOLE undump.obj
; dir undump.exe

.586
.model flat, stdcall
option casemap:none


include kernel32.inc
include windows.inc
includelib kernel32.lib

err_badhexdigit equ 1
err_openquotes equ 2

preserve	TEXTEQU	<edi esi>	; macros for routines
POINTER	TEXTEQU	<PTR>	; that with prototypes

snglquote equ 27h
dblquote equ 22h

@jmpshortifquote macro whereto
    cmp al,dblquote
    je short whereto
    cmp al,snglquote
    je short whereto
endm

@jmpshortifspace macro whereto
    cmp al,20h
    je short whereto
    cmp al,9
    je short whereto
endm

.data
ExitAddress dd ?
OutputBuffer dd ?
OutputLength dd ?
StdOut dd ?
StdErr dd ?
AmtWritten dd ?   ;for kernel32.WriteFile
ErrorHandlers dd Error1,Error2

ReadD1	PROTO dst : POINTER, src : POINTER, len : POINTER

.code
Start:
    pop ExitAddress
    sub esp,100h
    mov OutputBuffer,esp
    invoke GetStdHandle, STD_INPUT_HANDLE
    mov StdErr, eax
    invoke GetStdHandle, STD_OUTPUT_HANDLE
    mov StdOut, eax
    call GetCommandLine
    xchg eax, edi
	mov	al, '"'	; get args
	cmp	byte ptr [edi], al
	jne @F
	mov	cx, 100h
	inc	edi
	repne scasb
    inc edi
@@: mov	esi, edi
@@: lodsb
    call ZF_if_letter
    jz @B
    dec esi
    invoke ReadD1, OutputBuffer, esi, offset OutputLength
    jc HandleError
    invoke WriteFile,StdOut,OutputBuffer,OutputLength,addr AmtWritten,0
	xor	eax, eax
    jmp ExitAddress
HandleError:
    jmp ErrorHandlers[4*eax-4]

Error1:        ;stubs
    jmp ExitAddress
Error2:
    jmp ExitAddress

;;;;;;;;;;;;;;;;;;;;;;; ReadD1 ;;;;;;;;;;;;;;;;;;;;;;;;
;input is three dwords:
;lpsz destination (lowest on the stack)
;lpsz source
;lpdword where to store the output length (highest on the stack), or null

;returns:
;if bad input, CF with eax=error code
;if success, NC with eax=output length >= 0

ReadD1	PROC USES preserve,
 dst : POINTER, src : POINTER, len : POINTER
    mov esi,src
    mov edi,dst
    mov ebx,RD1_firstdigit   ;ebx maintains a code address
        ;for handling for the next input character.
        ;The three possibilities are RD1_firstdigit,
        ;RD1_seconddigit, and RD1_inquotes.
ReadD1_loop:
    mov al,[esi]   ;lodsb is slower but okay
    add esi,1
    or al,al
    call ebx   ;all the different handlers ebx will not return if al is 0
    jmp short ReadD1_loop

RD1_openquotes:
    push err_openquotes
RD1_baddata:      ;error code and return address are on the stack
    pop eax
    pop edi
    stc
    jmp RD1_pops
RD1_baddigit:
    push err_badhexdigit
    jmp RD1_baddata

RD1_store_ah_pop_finish:
    mov al,ah
    stosb
RD1_pop_finish:
    pop eax
RD1_finish:
    mov byte ptr[edi],0
    xchg eax,edi
    sub eax,dst
    mov edi,len
    test edi,edi     ;clc
    jz RD1_pops
    stosd
	jmp RD1_pops

RD1_inquotes:
    jz RD1_openquotes
    cmp al,dl
    je short RD1_reset
    stosb
    retn

RD1_firstdigit:
    jz RD1_pop_finish
    @jmpshortifquote RD1_startquote
    @jmpshortifspace RD1_ret
    call TranslateNibble
    jnz RD1_baddigit
    mov ah,al
    mov ebx,RD1_seconddigit
RD1_ret:
    retn

RD1_hangingdigit:
    mov [edi],ah
    inc edi
RD1_startquote:
    mov dl,al
    mov ebx,RD1_inquotes
    retn
RD1_seconddigit:
    jz RD1_store_ah_pop_finish
    @jmpshortifspace RD1_sto_ah
    @jmpshortifquote RD1_hangingdigit
    call TranslateNibble
    jnz RD1_baddigit
    shl ah,4
    or ah,al
RD1_sto_ah:
    mov al,ah
    stosb
RD1_reset:
    mov ebx,RD1_firstdigit
    retn

TranslateNibble:  ;ascii hex digit in al
    cmp al,"0"    ;returns al=nibble with ZF if a valid hex digit
    jb short TN_ret  ;returns NZ if an invalid digit
    cmp al,"9"
    jbe short TN_0_to_9
    or al,20h    ;al should now be a-f
    sub al,"a"
    jb short TN_ret   ;with NZ
    cmp al,5
    ja short TN_ret   ;with NZ
    add al,0Ah
TN_set_ZF:
    cmp al,al
TN_ret:
    retn
TN_0_to_9:
    sub al,"0"
    jmp short TN_set_ZF

RD1_pops:
    ret
ReadD1	ENDP

ZF_if_letter:
	cmp	al, '.'
	je short @F
    mov ah,al
    or ah,20h    ;convert to lower case, if it is a letter
    cmp ah,"a"
    jb short @F
    cmp ah,"z"
    ja short @F
    and ah,0
@@: ret

end Start
